unit MultiMtr;

{
  Naam:    TMultiMeter
  Versie:  1.0
  Datum:   30 oktober 2000

  Deze unit bevat een component voor het uitlezen
  van de meest voorkomende Multimeters. Voor dit component is
  TCommPortDriver voor Delphi 2 vereist. Deze werkt ook onder Delphi 5.

  TMultiMeter is een onderdeel van AdLentis Lia Controls

  Voor meer info over het gebruik van
  TMultiMeter, zie het bestand MultiMtr.txt

  Copyright (c) 1996-98 door Marco Cocco. Alle rechten voorbehouden.
  Copyright (c) 1996-98 door d3k Software Company. Alle rechten voorbehouden.
  Copyright (c) 2000 door Adriaan van Lent (AdLentis). Alle rechten voorbehouden.

  Web:    www.adlentis.net
  E-mail: webmaster@adlentis.net

  Het gebruiken, kopiren, aanpassen en gratis verspreiden van deze
  software met documentatie voor welk doel dan ook is bij deze
  toegestaan onder de voorwaarde dat bovenstaande copyright
  vermeldingen en deze toestemming voorkomen in alle ondersteunende
  documentatie.

}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, CommDrv;

type
  TMeterMode = (mmUnknown, mmDCVolt, mmACVolt, mmDCAmp, mmACAmp,
    mmFreq, mmResist, mmDiode);
  TMeterError = (meNone, meInvalid, meOverLimit, meTimeOut);
  TMeterErrorEvent = procedure(Sender: TObject; Error: TMeterError) of object;
  TMultiMeter = class(TCustomPanel)
  private
    FCPDrv: TCommPortDriver;   // Com poort interface
    FData: String;             // Ontvangen, nog niet verwerkte data
    FValue: Real;              // Laatste waarde
    FUnit: String;             // Meet eenheid
    FMode: TMeterMode;         // Meet modus
    FError: TMeterError;       // Error status
    FInterval: Cardinal;       // Meet interval
    FTimeout: Cardinal;        // Timeout
    FRequestTime: TDateTime;   // Laatste dataverzoek
    FRequestBusy: Boolean;     // Bezig met wachten op data
    FOverLimitText: String;    // Display-melding bij OL
    FInvalidText: String;      // Display-melding bij onbekende data
    FTimeOutText: String;      // Display-melding bij timeout
    FOffLineText: String;      // Display-melding bij geen verbinding
    FShowMode: Boolean;        // Toon meetmodus-afkorting in display?
    FNotifyWnd: HWND;          // Timer voor sturen dataverzoek
    FConnected: Boolean;       // COM poort open en errorstatus bekend?
    FOnConnect: TNotifyEvent;  // Wordt geactiveerd na COM poort open en error bekend 
    FOnError: TMeterErrorEvent;// Wordt geactiveerd bij errorverandering (na OnConnect)
    function GetPort: TPortNumber;
    procedure SetPort(Value: TPortNumber);
    function GetConnected: Boolean;
    function GetValue: Real;
    procedure RcvData(Sender: TObject; DataPtr: Pointer;
      DataSize: Integer);
    procedure SetOffLineText(Value: String);
    procedure SetInvalidText(Value: String);
    procedure SetTimeoutText(Value: String);
    procedure SetOverLimitText(Value: String);
    procedure SetInterval(Value: Cardinal);
    procedure TimerWndProc(var Msg: TMessage);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Connect: Boolean;
    procedure Disconnect;
    property MeterMode: TMeterMode read FMode;
    property ValueUnit: String read FUnit;
    property Value: Real read GetValue;
    property Error: TMeterError read FError;
    property Connected: Boolean read GetConnected;
  published
    // New properties
    property Port: TPortNumber read GetPort write SetPort;
    property ErrOverLimitText: String read FOverLimitText write SetOverLimitText;
    property ErrInvalidText: String read FInvalidText write SetInvalidText;
    property ErrTimeOutText: String read FTimeOutText write SetTimeOutText;
    property ErrOffLineText: String read FOffLineText write SetOffLineText;
    property ShowMode: Boolean read FShowMode write FShowMode;
    property Interval: Cardinal read FInterval write SetInterval;
    property Timeout: Cardinal read FTimeout write FTimeout;
    property OnError: TMeterErrorEvent read FOnError write FOnError;
    property OnConnect: TNotifyEvent read FOnConnect write FOnConnect;
    // Inherited properties
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BiDiMode;
    property BorderWidth;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property UseDockManager default True;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FullRepaint;
    property Font;
    property Locked;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

procedure Register;

implementation

var
  OldData: String;

function ConvertValue(V: Real; C: Char): Real;
begin
  Case C of
    'a' : V := V * 0.000000000000000001;
    'f' : V := V * 0.000000000000001;
    'p' : V := V * 0.000000000001;
    'n' : V := V * 0.000000001;
    'u' : V := V * 0.000001;
    'm' : V := V * 0.001;
    'c' : V := V * 0.01;
    'd' : V := V * 0.1;
    'k' : V := V * 1000;
    'K' : V := V * 1000;
    'M' : V := V * 1000000;
    'G' : V := V * 1000000000;
    'T' : V := V * 1000000000000;
    'P' : V := V * 1000000000000000;
    'E' : V := V * 1000000000000000000;
  end;
  Result := V;
end;

(********************** TMultiMeter ***********************)

constructor TMultiMeter.Create(AOwner: TComponent);
begin
  Inherited Create(AOwner);
  FCPDrv := TCommPortDriver.Create(Self);
  // Protocol instellingen
  With FCPDrv do begin
    BaudRate      := br1200;
    DataBits      := db7BITS;
    StopBits      := sb2BITS;
    Parity        := ptNONE;
    SwHandshaking := shNONE;
    HwHandshaking := hhNONE;
    OnReceiveData := RcvData;
  end;
  FOverLimitText := 'OL';
  FInvalidText   := 'INVALID';
  FTimeOutText   := 'TIMEOUT';
  FOffLineText   := 'OFFLINE';
  Caption := FOffLineText;
  FTimeout := 2000;
  FInterval := 1000;
  FRequestTime := 0;
  FRequestBusy := False;
  If not (csDesigning in ComponentState) then
    FNotifyWnd := AllocateHWnd(TimerWndProc);
end;

destructor TMultiMeter.Destroy;
begin
  DeallocateHWnd(FNotifyWnd);
  FCPDrv.Free;
  Inherited Destroy;
end;

procedure TMultiMeter.RcvData(Sender: TObject;
  DataPtr: Pointer; DataSize: Integer);
var
  P: PChar;
  I: Integer;
  B: Boolean;
  S, sData, sMode, sVal, sUnit: String;
  TmpValue: Real;
  TmpUnit : String;
  TmpMode : TMeterMode;
  TmpError: TMeterError;
  OldDecSep: Char;
begin

  If not FConnected then begin
    FConnected := True;
    If Assigned(FOnConnect) then FOnConnect(Self);
  end;

  TmpError := FError;

  // Verwerk data...
  P := DataPtr;
  S := '';
  For I := 1 to DataSize do begin
    S := S + P^;
    Inc(P);
  end;
  S := OldData + S;

  // Nieuwe gegevens?
  I := Pos(#13,S);
  sData := '';
  While I > 0 do begin
    sData := Copy(S,1,I-1);
    Delete(S,1,I+1);
    I := Pos(#13,S);
  end;
  OldData := S;

  TmpMode := mmUnknown;

  BEGIN
    // Verwerk data...
    If Length(sData) = 13 then begin

      FRequestBusy := False;
      TmpError := meNone;

      // Formaat: zie Metex.txt
      // Splits data
      sMode := Trim(Copy(sData,1,2));
      sVal  := Trim(Copy(sData,4,6));
      sUnit := Trim(Copy(sData,10,4));
      // Bepaal waarde
        OldDecSep := DecimalSeparator;
        DecimalSeparator := '.';
        Try
          TmpValue := StrToFloat(sVal);
          If TmpValue = 0 then TmpValue := Abs(TmpValue);
        except
          TmpValue := 0;
          S := StringReplace(sVal,'.','',[rfReplaceAll]);
          // OVER LIMIET
          If Pos('OL',UpperCase(S)) > 0 then
            TmpError := meOverLimit
          // ONGELDIG
          else
            TmpError := meInvalid;
        end;
        DecimalSeparator := OldDecSep;
        // Reken waarde om
        If sMode = 'DC' then begin
          Case sUnit[Length(sUnit)] of
            'V': begin
                   TmpMode := mmDCVolt; // Gelijkspanning
                   TmpUnit := 'V';
                   If Length(sUnit) >= 2 then
                     TmpValue := ConvertValue(TmpValue,sUnit[Length(sUnit)-1]);
                 end;
            'A': begin
                   TmpMode := mmDCAmp; // Gelijkstroom
                   TmpUnit := 'A';
                    If Length(sUnit) >= 2 then
                    TmpValue := ConvertValue(TmpValue,sUnit[Length(sUnit)-1]);
                 end;
          end; // Case
        end else
        If sMode = 'AC' then begin
          Case sUnit[Length(sUnit)] of
            'V': begin
                   TmpMode := mmACVolt; // Wisselspanning
                   TmpUnit := 'V';
                   If Length(sUnit) = 2 then
                     TmpValue := ConvertValue(TmpValue,sUnit[1]);
                 end;
            'A': begin
                   TmpMode := mmACAmp; // Wisselstroom
                   TmpUnit := 'A';
                   If Length(sUnit) = 2 then
                     TmpValue := ConvertValue(TmpValue,sUnit[1]);
                 end;
          end; // Case
        end else
        If sMode = 'DI' then begin
          TmpMode := mmDiode; // Diodespanning
          TmpUnit := 'V';
          If Length(sUnit) = 2 then
            TmpValue := ConvertValue(TmpValue,sUnit[1]);
        end else
        If sMode = 'FR' then begin
          TmpMode := mmFreq; // Frequentie
          TmpUnit := 'Hz';
          If Length(sUnit) = 2 then
            TmpValue := ConvertValue(TmpValue,sUnit[1]);
        end else
        If sMode = 'OH' then begin
          TmpMode := mmResist; // Weerstand
          TmpUnit := 'Ohm';
          If Length(sUnit) = 4 then
            TmpValue := ConvertValue(TmpValue,sUnit[1]);
        end else
        begin
          TmpMode := mmUnknown;
          TmpUnit := '';
          TmpError := meInvalid;
        end;

        // WAARDE DOORVOEREN
        FValue := TmpValue;
        FMode  := TmpMode;
        FUnit  := TmpUnit;
        If FError <> TmpError then begin
          FError := TmpError;
          If Assigned(FOnError) then FOnError(Self,FError);
        end;

      // Bepaal panelcaption
      If TmpError = meNone then begin
        S := Trim(sVal);
        // Overbodige nullen weghalen
        B := S[1] = '-';
        If B then Delete(S,1,1);
        While (S[1]='0') and (S[2]<>'.') do begin
          Delete(S,1,1);
          S := S + '0';
        end;
        If B then S := '-'+S;
        // Overbodig minteken weghalen
        If S='-0.000' then S:='0.000';
        // Aanvullen met eenheid
        S := S + ' ' + sUnit;
        If FShowMode then S := sMode + ' ' + S;
      end else

      // Not valid:
      begin
        Case TmpError of
          meOverLimit:
            S := FOverLimitText;
          meInvalid:
            S := FInvalidText;
        end; // Case
      end;

      If S <> Caption then Caption := Trim(S);

    end; // If length(sData) = 13
    
  END;

end;

function TMultiMeter.GetValue: Real;
begin
  Result := 0;
  If not Connected then raise Exception.Create('Not connected.');
  Case FError of
    meNone:
      Result := FValue;
    meInvalid:
      raise Exception.Create('Invalid value.');
    meOverLimit:
      raise Exception.Create('Over limit.');
    meTimeOut:
      raise Exception.Create('Timeout.');
  end; // Case
end;

procedure TMultiMeter.SetPort(Value: TPortNumber);
begin
  // Verandering doorvoeren...
  FCPDrv.Port := Value;
end;

procedure TMultiMeter.Disconnect;
var
  I: Integer;
begin
  If not FCPDrv.Connected then Exit;

  FConnected := False;
  // Stop timer...
  KillTimer(FNotifyWnd,1);
  // Disconnect
  FCPDrv.Disconnect;
  Caption := FOffLineText;
  // Discard unprocessed data...
  OldData := '';
  // Reset
  FValue := 0;
  FError := meNone;
  FRequestBusy := False;
end;

function TMultiMeter.Connect: Boolean;
begin
  Result := True;
  If FCPDrv.Connected then Exit;
  Result := FCPDrv.Connect;
  Caption := FOffLineText;
  If Result then begin
    // Start timer!
    FCPDrv.SendString('D');
    SetTimer(FNotifyWnd,1,FInterval,nil);
  end;  
end;

function TMultiMeter.GetPort: TPortNumber;
begin
  Result := FCPDrv.Port;
end;

function TMultiMeter.GetConnected: Boolean;
begin
  Result := FConnected;
end;

procedure TMultiMeter.SetOffLineText(Value: String);
begin
  If FOffLineText <> Value then begin
    FOffLineText := Value;
    If not FCPDrv.Connected then Caption := Value;
  end;
end;

procedure TMultiMeter.SetInvalidText(Value: String);
begin
  If FInvalidText <> Value then begin
    FInvalidText := Value;
    If FError = meInvalid then Caption := Value; 
  end;
end;

procedure TMultiMeter.SetTimeoutText(Value: String);
begin
  If FTimeoutText <> Value then begin
    FTimeoutText := Value;
    If FError = meTimeout then Caption := Value;
  end;
end;

procedure TMultiMeter.SetOverLimitText(Value: String);
begin
  If FOverLimitText <> Value then begin
    FOverLimitText := Value;
    If Error = meOverLimit then Caption := Value;
  end;
end;

procedure TMultiMeter.SetInterval(Value: Cardinal);
begin
  If Value <> FInterval then begin
    If Connected then
      KillTimer(FNotifyWnd,1);
    FInterval := Value;
    If Connected then
      SetTimer(FNotifyWnd,1,FInterval,Nil);
  end;
end;

procedure TMultiMeter.TimerWndProc(var Msg: TMessage);
begin
  If (msg.Msg = WM_TIMER) and FCPDrv.Connected then
  begin
    // Controleer op time-out
    If (FRequestBusy = True) then
    begin
      If (Now >= FRequestTime + (FTimeout/1000/SecsPerDay)) then
      begin
        If FError <> meTimeout then
        begin
          FError := meTimeout;
          // Pas verbonden...
          If not FConnected then begin
            FConnected := True;
            If Assigned(FOnConnect) then FOnConnect(Self);
          end;
          If Assigned(FOnError) then FOnError(Self,FError);
        end;
        Caption := FTimeOutText;
        FRequestBusy := False;
      end;
    end else begin
      // Stuur datarequest
      FCPDrv.SendString('D');
      FRequestTime := Now;
      FRequestBusy := True;
    end;
  end;
end;

(***********************************************)

procedure Register;
begin
  RegisterComponents('System', [TMultiMeter]);
end;

end.
